use crate::queue::Message;
use std::{
    sync::mpsc::{channel, Receiver, Sender},
    thread::spawn,
};

#[derive(Clone)]
pub enum BackEndMessage {
    Get,
    Response(String),
}

struct BackEnd {
    tx: Sender<Message>,
    rx: Receiver<Message>,
}

impl BackEnd {
    fn new(tx: Sender<Message>, rx: Receiver<Message>) -> Self {
        Self { tx: tx, rx: rx }
    }

    fn start(queue_tx: Sender<Message>) {
        spawn(move || {
            let (tx, rx) = channel();
            let service = Self::new(queue_tx.clone(), rx);
            queue_tx.send(tx.clone().into()).unwrap();
            service.listen();
        });
    }

    fn listen(&self) {
        loop {
            self.rx.recv().unwrap();
            self.tx
                .send(BackEndMessage::Response("Something goes here".to_string()).into())
                .unwrap();
        }
    }
}

#[cfg(test)]
mod backends {
    use super::*;
    use std::time::Duration;

    fn run_service() -> (Sender<Message>, Receiver<Message>) {
        let (tx, rx) = channel();
        let service_tx: Sender<Message>;
        BackEnd::start(tx);
        match rx.recv().unwrap() {
            Message::Register(result) => service_tx = result,
            _ => unreachable!("should register the service"),
        }
        (service_tx, rx)
    }

    #[test]
    fn get_registered() {
        let (tx, rx) = channel();
        BackEnd::start(tx);
        match rx.recv_timeout(Duration::from_millis(500)).unwrap() {
            Message::Register(_) => {}
            _ => unreachable!("should register the service"),
        }
    }

    #[test]
    fn request_home_page() {
        let (tx, rx) = run_service();
        tx.send(BackEndMessage::Get.into()).unwrap();
        match rx.recv_timeout(Duration::from_millis(500)).unwrap() {
            Message::BEMsg(result) => match result {
                BackEndMessage::Response(_) => {}
                _ => unreachable!("wrong message type"),
            },
            _ => unreachable!("wrong message type"),
        }
    }
}
